/*
 * Copyright (c) 2012 - 2013 Alexandre Tuleu
 *
 * This file is part of biorob-cpp.
 *
 *   biorob-cpp is free software: you can redistribute it and/or modify
 *   it under the terms of the GNU Lesser General Public License as
 *   published by the Free Software Foundation, either version 3 of the
 *   License, or (at your option) any later version.
 *
 *   biorob-cpp is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of both the GNU General Public
 *   License AND GNU Lesser General Public License along with
 *   biorob-cpp.  If not, see <http://www.gnu.org/licenses/>.
 */


/*
 * Copyright (c) 2012  Alexandre Tuleu
 *
 * This file is part of biorob-cpp.
 *
 *   biorob-cpp is free software: you can redistribute it and/or modify
 *   it under the terms of the GNU Lesser General Public License as
 *   published by the Free Software Foundation, either version 3 of the
 *   License, or (at your option) any later version.
 *
 *   biorob-cpp is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of both the GNU General Public
 *   License AND GNU Lesser General Public License along with
 *   PROJECT.  If not, see <http://www.gnu.org/licenses/>.
 */


/**
 * \file Interface.cpp
 *
 *  \date Oct 24, 2012
 *  \author Alexandre Tuleu
 */

#include "Interface.h"
#include <cmath>
#include "signal.h"
#include <cstdlib>

namespace cli {

	bool Interface::s_alreadyExists = false;
	bool Interface::s_hasColor = false;
	bool Interface::s_ncursesInitialized(false);
	const unsigned int Interface::MinWidgetSize = 4;
	const unsigned int Interface::MinPromptHeight = 5;
	const unsigned int Interface::MaxPromptHeight = 10;

	std::streambuf * Interface::s_coutBuf(NULL);
	std::streambuf * Interface::s_cerrBuf(NULL);
	std::streambuf * Interface::s_clogBuf(NULL);

	bool Interface::HasColor(){
		return s_hasColor;
	}

	Interface::Interface(unsigned int rows, unsigned int cols, bool handleSignals )
		: d_widgetHeight(0)
		, d_widgetWidth(0)
		, d_promptHeight(MinPromptHeight)
		, d_displayHeight(0)
		, d_rows(rows)
		, d_cols(cols)
		, d_prompt()
		, d_generalDisplay()
		, d_variables()
		, d_commands()
		, d_userAskToQuit(false)
		, d_promptString(">>> "){
		if(s_alreadyExists){
			throw std::logic_error("Could only have one instance of Interface");
		}

		if(handleSignals){
			signal(SIGINT,&SignalDeinitFunction);
		}

		InitializeNCurses();
		try {
			InitializeSizes();
			s_alreadyExists = true;
			Widget::Def promptDef,displayDef;
			promptDef.height = (int)d_promptHeight;
			promptDef.width = COLS;
			promptDef.y = (int) (d_rows) * (int) (d_widgetHeight)
			        + (int) (d_displayHeight);
			promptDef.x = 0;
			promptDef.title = "Command Line";
			displayDef.height = (int)d_displayHeight;
			displayDef.width = COLS;
			displayDef.y = (int) (d_rows) * (int) (d_widgetHeight);
			displayDef.x = 0;
			displayDef.title = "Program Output";
			refresh();
			d_generalDisplay = std::tr1::shared_ptr < DisplayWidget> (new DisplayWidget(displayDef));

			d_prompt = std::tr1::shared_ptr < Prompt > (new Prompt(promptDef));
			s_coutBuf = std::cout.rdbuf(d_generalDisplay->Stream().rdbuf());
			s_cerrBuf = std::cerr.rdbuf(d_generalDisplay->ErrStream().rdbuf());
			s_clogBuf = std::clog.rdbuf(d_generalDisplay->ErrStream().rdbuf());

			d_prompt->AddCompletion("exit");
			d_prompt->AddCompletion("quit");
		} catch (std::exception const& e) {
			DeInitializeNCurses();
			throw ;
		}
	}

	Interface::~Interface(){
		DeInitializeNCurses();
	}


	void Interface::SignalDeinitFunction(int){
		DeInitializeNCurses();
		exit(0);
	}

	void Interface::DeInitializeNCurses(){
		if(s_ncursesInitialized){
			endwin();
			s_ncursesInitialized = false;
		}
		if(s_coutBuf) {
			std::cout.rdbuf(s_coutBuf);
		}
		if(s_cerrBuf) {
			std::cerr.rdbuf(s_cerrBuf);
		}
		if(s_clogBuf) {
			std::clog.rdbuf(s_clogBuf);
		}
	}

	void Interface::InitializeNCurses(){
		if(s_ncursesInitialized){
			return;
		}
		initscr();
		s_ncursesInitialized = true;
		keypad(stdscr, true);
		cbreak();
		noecho();
		if (has_colors()) {
			s_hasColor = true;
			if(ERR == start_color()){
				endwin();
				throw std::runtime_error("Could not init color.");
			}
			if( ERR == init_pair(WhiteOnBlack,COLOR_WHITE,COLOR_BLACK)){
				endwin();
				throw std::runtime_error("Could not init color pair White on black");
			}
			if(ERR == init_pair(GreenOnBlack,COLOR_GREEN,COLOR_BLACK)){
				endwin();
				throw std::runtime_error("Could not init color pair Green on black");
			}
			if( ERR == init_pair(RedOnBlack,COLOR_RED,COLOR_BLACK)){
				endwin();
				throw std::runtime_error("Could not init color pair red on black");
			}
			attrset(COLOR_PAIR(WhiteOnBlack));
			refresh();
		} else {
			s_hasColor = false;
		}
	}
	const Interface::VariableByName & Interface::Variables() const{
		return d_variables;
	}
	void Interface::AddVariable(VariableBase::Ptr var){
		d_variables[var->Name()] = var;
		d_prompt->AddCompletion(var->Name());
	}

	void Interface::AddCommand(Command::Ptr command){
		d_commands[command->Name()] = command;
		d_prompt->AddCompletion(command->Name());
	}

	void Interface::RemoveCommand(const std::string & name){
		CommandByName::iterator fi = d_commands.find(name);
		if(fi == d_commands.end()){
			return;
		}
		d_commands.erase(fi);
		VariableByName::const_iterator vi = d_variables.find(name);
		if(vi == d_variables.end()){
			d_prompt->RemoveCompletion(name);
		}
	}
	void Interface::RemoveVariable(VariableBase::Ptr var){
		VariableByName::iterator fi = d_variables.find(var->Name());
		if(fi == d_variables.end()){
			return;
		}
		d_variables.erase(fi);
		CommandByName::const_iterator ci = d_commands.find(var->Name());
		if(ci == d_commands.end()){
			d_prompt->RemoveCompletion(var->Name());
		}
	}

	bool Interface::Process(){
		std::vector<std::string> tokens;
		std::string input = d_prompt->GetLine(d_promptString);

		//tokenizing by spaces
		SplitInToken(input,tokens);

		if(tokens.empty()){
			return UserAskedToQuit();
		}

		//Check builtin exit or quit command.
		if(tokens.front() == "exit" || tokens.front()== "quit"){
			if (tokens.size() > 1){
				d_generalDisplay->ErrStream() << "<<< exit command don't take any arguments"
				                           << std::endl;
			} else {
				d_userAskToQuit = true;
			}
			return UserAskedToQuit();
		}

		//Now check for user built-in command

		CommandByName::iterator cmd = d_commands.find(tokens.front());


		if(cmd != d_commands.end()){
			CallCommand(cmd,tokens);
			return UserAskedToQuit();
		}

		VariableByName::iterator var = d_variables.find(tokens.front());

		if(var != d_variables.end()){
			ProcessVariable(var,tokens);
			return UserAskedToQuit();
		}

		d_generalDisplay->ErrStream() << "<<< Undefined variable or command '" << tokens.front()
		                           << "'." << std::endl;


		return UserAskedToQuit();
	}

	bool Interface::UserAskedToQuit(){
		return d_userAskToQuit;
	}

	void Interface::AddBinding(int c,KeyCallback::Ptr callback){
		d_prompt->BindKey(c,callback);
	}
	void Interface::Unbind(int c){
		d_prompt->UnBindKey(c);
	}


	void Interface::InitializeSizes(){
		if(LINES <= (int)MinPromptHeight){
			std::ostringstream os;
			os << "Your terminal is just too small " << LINES << "x" << COLS;
			throw std::runtime_error(os.str());
		}
		d_widgetHeight = std::floor((float) (LINES) / (d_rows + 2));
		d_promptHeight = d_widgetHeight;
		if (d_promptHeight < MinPromptHeight) {
			d_widgetHeight = std::floor(
			        ((float) (LINES) - MinPromptHeight) / (d_rows + 1));
			d_promptHeight = MinPromptHeight;
		}

		if(d_promptHeight > MaxPromptHeight){
			d_promptHeight = MaxPromptHeight;
			d_widgetHeight = std::floor(((float)LINES - MaxPromptHeight) / (d_rows + 1));
		}

		if(d_cols > 0){
			d_widgetWidth = std::floor((float) (COLS) / d_cols);
		} else {
			d_widgetWidth = (unsigned int)COLS;
		}
		if(d_widgetHeight < MinWidgetSize || d_widgetWidth < MinWidgetSize){
			std::ostringstream os;
			os << "Could not arrange grid layout " << d_rows << "x" << d_cols
			   << " because the screen is too small : " << LINES << "x" << COLS;
			throw std::runtime_error(os.str());
		}

		d_displayHeight = (unsigned int)LINES - d_rows * d_widgetHeight - d_promptHeight;


	}


	std::ostream & Interface::IOut(){
		return d_generalDisplay->Stream();
	}

	std::ostream & Interface::IErr(){
		return d_generalDisplay->ErrStream();
	}

	const std::string Interface::SpecialChars("=,()$*/<>&|");
	const std::string Interface::SpaceChars(" \t\v\n\f\r");

	void Interface::SplitInToken(const std::string & input,
	                             std::vector<std::string> & tokens) const{
		tokens.clear();

		std::string::size_type index(0),startToken,endToken;
		while(index < input.size()){
			startToken = input.find_first_not_of(SpaceChars,index);
			if(startToken == std::string::npos){ //reach end of string.
				break;
			}
			//test if special char
			if(SpecialChars.find(input.at(startToken)) != std::string::npos){
				endToken = startToken + 1;
			} else {
				endToken = input.find_first_of(SpaceChars + SpecialChars, startToken);
			}
			if(endToken == std::string::npos){
				tokens.push_back(input.substr(startToken));
			} else {
				tokens.push_back(input.substr(startToken,endToken - startToken));
			}
			if(endToken >= input.size()){
				break;
			}
			index = endToken;

		}
	}

	void Interface::CallCommand(CommandByName::iterator & cmd,
	                            std::vector<std::string> & tokens){
		Command::ListOfArguments args;
		if(tokens.size() > 1){
			//Extract args
			if(!ExtractArgs(tokens,args)){
				return;
			}
		}

		unsigned int nbMandatoryArgs = cmd->second->NumberbOfArguments();

		if((args.size() < nbMandatoryArgs ) ||
				(args.size() > nbMandatoryArgs && !cmd->second->AcceptOptionnalArgs())){
			d_generalDisplay->ErrStream() << ">>> Wrong number of arguments for command '"
			                              << cmd->second->Name() << "' expected "
			                              << nbMandatoryArgs << " got "
			                              << args.size() << "." << std::endl;
			return;
		}

		//Now perform the actual Call
		if(!cmd->second->Call(args)) {
			d_generalDisplay->ErrStream() << "<<< Unable to perform '"
			                              << cmd->second->Name() << "'."
			                              << std::endl;
		}

	}

	bool Interface::ExtractArgs(std::vector<std::string> & tokens,
	                            Command::ListOfArguments & args){
		using namespace std;
		if (tokens.size() == 1){
			return true;
		}
		vector<string>::size_type start(1);
		if(tokens.at(1) == "("){
			if(tokens.back() != ")"){
				d_generalDisplay->ErrStream() << "<<< Mismatching parenthesis in expression." << endl;
				return false;
			}
			tokens.pop_back();
			++start;
		}
		string::size_type spChar;

		for(vector<string>::const_iterator t = tokens.begin() + (std::vector<string>::difference_type)start;
				t != tokens.end();
				++t){

			spChar = t->find_first_of(SpecialChars);

			if( spChar != string::npos){
				d_generalDisplay->ErrStream() << "<<< Expected one argument name, got '"
				                              << t->at(spChar) <<"'." << endl;
				return false;
			}

			//its a name, replace immediatly variable if it's one.

			VariableByName::const_iterator var = d_variables.find(*t);
			if(var == d_variables.end()){ //instantiate a temporary string variable
				ostringstream os;
				os << "argv" << args.size();
				args.push_back(VariableBase::Ptr(new Variable<string>(os.str(),*t)));
			} else {
				args.push_back(var->second);
			}

			//check for separator.
			++t;

			if(t == tokens.end()){//last token reached.
				break;
			}

			if(*t != ","){
				d_generalDisplay->ErrStream() << "<<< Bad command call, expected ',' argument separator, got '"
				                              << *t << "'." << endl;
				return false;
			}
		}


		return true;

	}

	void Interface::ProcessVariable(VariableByName::iterator & var,
	                                std::vector<std::string> & tokens){
		if(tokens.size() == 1){
			//normal out
			d_generalDisplay->Stream() << "<<< " << var->second->Name() << " = "
			                           << var->second->toString() << std::endl;
			return;
		}

		if(tokens.size() >= 2 && tokens[1] == "="){
			if(tokens.size() != 3){
				d_generalDisplay->ErrStream() << "<<< Setting a variable could only take one arguments "
				                              << "after '='." << std::endl;
				return;
			}
			//search if it is a variable name;
			VariableByName::const_iterator var2 = d_variables.find(tokens[2]);
			std::string value = var2 == d_variables.end() ? tokens[2] : var2->second->toString();
			if(!var->second->SetFromString(value)){
				d_generalDisplay->ErrStream() << "Unable to set " << var->second->Name() << " to ";
				if(var2 == d_variables.end()){
					d_generalDisplay->ErrStream() << "'" << value << "'." << std::endl;
				} else {
					d_generalDisplay->ErrStream() << var2->second->Name() << " which is '" << value
					                              << "'." << std::endl;
				}
			} else {
				//normal out
				d_generalDisplay->Stream() << "<<< " << var->second->Name() << " <- "
				                           << var->second->toString() << std::endl;
			}

		} else {
			d_generalDisplay->ErrStream() << "<<< Unknown operation '" << tokens[1] << "' on "
			                              << var->second->Name() << "." << std::endl;
			return;
		}

	}

	const std::string & Interface::GetPrompt() const {
		return d_promptString;
	}

	void Interface::SetPrompt(const std::string & p){
		if(p.empty()){
			return;
		}
		d_promptString = p;
		if(p.at(p.size() -1) != ' '){
			d_promptString.append(" ");
		}
	}

} /* namespace cli */
