/*
 * Copyright (c) 2012 - 2013 Alexandre Tuleu
 *
 * This file is part of biorob-cpp.
 *
 *   biorob-cpp is free software: you can redistribute it and/or modify
 *   it under the terms of the GNU Lesser General Public License as
 *   published by the Free Software Foundation, either version 3 of the
 *   License, or (at your option) any later version.
 *
 *   biorob-cpp is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of both the GNU General Public
 *   License AND GNU Lesser General Public License along with
 *   biorob-cpp.  If not, see <http://www.gnu.org/licenses/>.
 */


/**
 * \file Logger.h
 *
 *  \date Jul 26, 2012
 *  \author Alexandre Tuleu
 */

#ifndef BIOROB_CPP_LOG_LOGGER_H_
#define BIOROB_CPP_LOG_LOGGER_H_

#include <iostream>

#include <biorob-cpp/base/common.h>



#ifndef NDEBUG
#define log_headers __FILE__, ":", __LINE__, " : "

#define log_raw_nh(l , ...) do {\
	loggers::l , __VA_ARGS__ ;\
}while(0)

#define log_raw(l , ...) do {\
	if(loggers::l.AppendName()){\
		loggers::l.Stream() << "[" << loggers::l.Name()<< "] ";\
	}\
	loggers::l , log_headers , __VA_ARGS__ ;\
}while(0)


#define log(l , ...) do {\
	log_raw(l , __VA_ARGS__ );\
	std::endl(loggers::l.Stream());\
}while(0)

#else
#define log_headers __FILE__, ":", __LINE__, " :"

#define log_raw_nh(l , ...) do { } while(0)

#define log_raw(l , ... ) do { } while(0)

#define log(l , ...) do{ } while(0)
#endif


/**
 * Logger class, super simple to use. Just use,
 *
 * log(warn, "my awesome log of my object", my_object);
 *
 * and it will log in warn.log file the line :
 * __FILE__:__LINE__ : my awesome lof of my object <my_object>
 *
 * any object that can be passed to a std::ostream can go in the logger.
 *
 * Tired to output to a file, try logger::warn.SetCerr(); or
 * logger::warn.SetCout(); or logger::warn.SetFile("foo.my_log");.
 * Even logger::warn.SetNull() do the trick !
 *
 * You have issue with the logging ) ? np ! define the preprocessor macro NDEBUG
 * and *all* logging information will be stripped away. Yes sir, no runtime
 * cost !
 *
 * and its less than 1k of line code. Enjoy !
 */
class Logger {
	DISABLE_COPY(Logger)
public:
	Logger(const std::string & name);
	virtual ~Logger();

	std::ostream & Stream();

	void SetNull();
	void SetFile(const std::string & filename);
	void SetClog();
	void SetCout();
	void SetCerr();

	const std::string & Name() const;
	bool AppendName() const;
private :
	void operator=(const Logger&); //UNDEFINED
	class NullStreamBuff : public std::streambuf{};

	void Clean();

	std::ostream * d_stream;
	static std::ostream s_nullStream;
	bool d_deleteStream;
	const std::string d_name;
};




template <typename T>
Logger & operator,(Logger & l , const T & t){
	l.Stream() << t;
	return l;
}


namespace loggers{
	static Logger trace("trace");
	static Logger debug("debug");
	static Logger info("info");
	static Logger warning("warning");
	static Logger error("error");
	static Logger fatal("fatal");
}

#endif // BIOROB_CPP_LOG_LOGGER_H_
