/**
 * \file Socket.h
 *
 *  \date Sep 18, 2013
 *  \author Alexandre Tuleu
 */

#ifndef BIOROB_CPP_NET_SOCKET_H_
#define BIOROB_CPP_NET_SOCKET_H_

#include <ios>
#include <stdint.h>
#include <stdexcept>
#include <tr1/memory>

#include <biorob-cpp/base/common.h>

namespace net {

class SocketFullIOError : public std::runtime_error {
public :
	enum IOType {
		READ = 0,
		WRITE
	};

	SocketFullIOError(IOType type,
	                  size_t total,
	                  size_t performed,
	                  unsigned int retries);
	virtual ~SocketFullIOError() throw();

	IOType Type() const;
	size_t TotalSize() const;
	size_t PerformedSize() const;
	size_t Retries() const;

private :
	static std::string Stringify(IOType type,
	                             size_t total,
	                             size_t performed,
	                             unsigned int retries);
	IOType d_type;
	size_t d_total;
	size_t d_performed;
	size_t d_retries;
};


class Socket {
	DISABLE_COPY(Socket);
	DISABLE_ASSIGNEMENT(Socket);
public:
	typedef std::tr1::shared_ptr<Socket> Ptr;
	Socket();
	Socket(int fd);
	virtual ~Socket();

	void SetReceiveFlags(int flags);
	void SetSendFlags(int flags);

	unsigned int GetMaxRetries();
	void SetMaxRetries(unsigned int maxRetries);

	size_t Read(uint8_t * buffer, size_t length);
	size_t Write(uint8_t * buffer, size_t length);

	void ReadAll(uint8_t * buffer, size_t length);
	void WriteAll(uint8_t * buffer, size_t length);

	void SetNonBlocking(bool block = true);

protected :
	static int GetSocketType(int fd);
	static int ConnectSocket(const std::string & name,
	                         unsigned int port,
	                         int type);

	void SetFd(int fd);


private :

	void Clean();

	int d_fd;
	int d_recvFlags;
	int d_sendFlags;
	unsigned int d_maxRetries;
};

} /* namespace net */
#endif /* BIOROB_CPP_SOCKET_H_ */
