/*
 * \file AvahiPublisher.cpp
 *
 *  \date Oct 18, 2013
 *  \author Alexandre Tuleu
 */

#include "Publisher.h"

#include <stdexcept>
#include <sstream>

#include <avahi-client/client.h>
#include <avahi-client/publish.h>

#include <avahi-common/alternative.h>
#include <avahi-common/simple-watch.h>
#include <avahi-common/malloc.h>
#include <avahi-common/error.h>
#include <avahi-common/timeval.h>

#include <biorob-cpp/log/Logger.h>
namespace net {
namespace dnssd {

// this publisher private implementation is based on the Avahi example.
// Just revamped to use C++ paradigm (exception and RAII)

class AvahiPublisher : public PublisherPrimpl {
public :
	AvahiPublisher(Publisher & publisher);
	virtual ~AvahiPublisher();


	void EnterMainLoop();
private :
	typedef std::tr1::shared_ptr< ::AvahiClient > ClientPtr;
	typedef std::tr1::shared_ptr< char > StrPtr;
	typedef std::tr1::shared_ptr<AvahiSimplePoll>     SimplePollPtr;

	class CollisionError {};

	void Clean();
	static AvahiStringList * ToAvahiStrList(const Publisher::ListOfTxtDataFields & txtData);
	static AvahiIfIndex      ToAvahiInterface(Interface i);
	static AvahiProtocol     ToAvahiProtocol(Protocol p);
	static void ClientCallback(AvahiClient * c,
	                           AvahiClientState state,
	                           void *userdata);
	static void EntryGroupCallback(AvahiEntryGroup *g,
	                               AvahiEntryGroupState state,
	                               void *userdata);

	void CreateServices(AvahiClient * c);
	void AddService(AvahiClient *c,
	                const ServiceEntry & entry);
	AvahiEntryGroup * d_group; //Is not shared_ptr as managed by client.
	SimplePollPtr d_poll;
	StrPtr        d_name;
};


PublisherPrimpl::Ptr PublisherPrimpl::Create(Publisher & publisher) {
	return PublisherPrimpl::Ptr(new AvahiPublisher(publisher));
}


AvahiPublisher::AvahiPublisher(Publisher & publisher)
	: PublisherPrimpl(publisher)
	, d_group(NULL){
}

AvahiPublisher::~AvahiPublisher() {
	Clean();
}

void AvahiPublisher::EnterMainLoop() {
	d_poll = SimplePollPtr(avahi_simple_poll_new(),&avahi_simple_poll_free);

	if(!d_poll) {
		throw std::runtime_error("avahi_simple_poll_new() failed");
	}

	d_name = StrPtr(avahi_strdup(GetPublisher().Name().c_str()),&avahi_free);
	int error;
	ClientPtr client(avahi_client_new(avahi_simple_poll_get(d_poll.get()),
	                                  (AvahiClientFlags)0,
	                                  &ClientCallback,
	                                  reinterpret_cast<void*>(this),
	                                  &error),
	                 &avahi_client_free);

	if(!client) {
		Clean();
		std::ostringstream os;
		os << "avahi_client_new() returned error ( " << error
		   << ": " << avahi_strerror(error) << ")";
		throw std::runtime_error(os.str());
	}

	error = avahi_simple_poll_loop(d_poll.get());
	// can return 1 on quit operation asked.
	if(error < 0) {
		Clean();
		std::ostringstream os;
		os << "avahi_simple_poll_loop() returned with error ( " << error
		   << ": " << avahi_strerror(error) << " )";
		throw std::runtime_error(os.str());
	}

}

void AvahiPublisher::Clean() {
	if(d_poll) {
		avahi_simple_poll_quit(d_poll.get());
	}
	d_poll  = SimplePollPtr();
	d_name  = StrPtr();
}

void AvahiPublisher::ClientCallback(AvahiClient * c,
                                    AvahiClientState state,
                                    void * userdata) {
	AvahiPublisher * itself = reinterpret_cast<AvahiPublisher*>(userdata);
	assert(c);

	switch (state) {
	case AVAHI_CLIENT_S_RUNNING:

		/* The server has startup successfully and registered its host
		 * name on the network, so it's time to create our services */
		itself->CreateServices(c);
		break;

	case AVAHI_CLIENT_FAILURE: {
		std::ostringstream os;
		int error = avahi_client_errno(c);
		os << "Had client failure ( " << error << ": " << avahi_strerror(error) << " )";
		avahi_simple_poll_quit(itself->d_poll.get());
		throw std::runtime_error(os.str());
	}
	case AVAHI_CLIENT_S_COLLISION:
		/* Let's drop our registered services. When the server is back
		 * in AVAHI_SERVER_RUNNING state we will register them
		 * again with the new host name. */
	case AVAHI_CLIENT_S_REGISTERING:
		/* The server records are now being established. This
		 * might be caused by a host name change. We need to wait
		 * for our own records to register until the host name is
		 * properly esatblished. */
		if (itself->d_group)
			avahi_entry_group_reset(itself->d_group);
		break;

	case AVAHI_CLIENT_CONNECTING:
		break;
	default :
		throw std::logic_error("Unexpected client state");
	}
}

void AvahiPublisher::CreateServices(AvahiClient * c) {
	assert(c);

	if(d_group == NULL) {
		d_group = avahi_entry_group_new(c,
		                                &EntryGroupCallback,
		                                reinterpret_cast<void*>(this));
	}

	if(d_group == NULL) {
		int error = avahi_client_errno(c);
		std::ostringstream os;
		os << "avahi_entry_group_new() returned error ( " << error << ": " << avahi_strerror(error) << " )";
		avahi_simple_poll_quit(d_poll.get());
		throw std::runtime_error(os.str());
	}

	if( ! avahi_entry_group_is_empty(d_group)) {
		return;
	}

	log(debug,"Adding service ", GetPublisher().Name(),".");

	for(ListOfServices::const_iterator service = Services().begin();
	    service != Services().end();
	    ++ service) {
		try {
			this->AddService(c,*service);
		} catch (const CollisionError & ) {
			d_name = StrPtr(avahi_alternative_service_name(d_name.get()),
			                &avahi_free);
			log(debug,"Got name collision renaming service to ",d_name.get());

			avahi_entry_group_reset(d_group);
			CreateServices(c);
			return;
		}

	}

	int res = avahi_entry_group_commit(d_group);
	if(res) {
		std::ostringstream os;
		os << "avahi_entry_group_commit() returned error ( " << res
		   << ": " << avahi_strerror(res) << " )";
		avahi_simple_poll_quit(d_poll.get());
		throw std::runtime_error(os.str());
	}
}



#define wrap_str(str,name) \
	const char * name  = str.empty() ? NULL : str.c_str()
void AvahiPublisher::AddService(AvahiClient * c,
                                const ServiceEntry & service) {
	assert(c);
	assert(d_group);

	wrap_str(service.type, type);
	wrap_str(service.domain, domain);
	wrap_str(service.host, host);

	int error = avahi_entry_group_add_service_strlst(d_group,
	                                                 ToAvahiInterface(service.interface),
	                                                 ToAvahiProtocol(service.protocol),
	                                                 (AvahiPublishFlags)0,
	                                                 d_name.get(),
	                                                 type,
	                                                 domain,
	                                                 host,
	                                                 service.port,
	                                                 ToAvahiStrList(service.txtData));
	log(debug,"Adds service ",d_name, 
	    " of type '" , type, 
	    "' on domain:host:port ", (domain ? domain : "<ANY_DOMAIN>") ,
	    ":",(host ? host : "<ANY_HOST>" ) ,
	    ":",service.port );

	if(error == AVAHI_ENTRY_GROUP_COLLISION) {
		log(debug,"Got error collision");
		throw CollisionError();
	}

	if(error) {
		std::ostringstream os;
		os << "avahi_entry_group_add_service_strlst() returned with error ( "
		   << error << ": " << avahi_strerror(error) << " )";
		avahi_simple_poll_quit(d_poll.get());
		throw std::runtime_error(os.str());
	}

	for(Publisher::ListOfSubtype::const_iterator subtype = service.subtypes.begin();
	    subtype != service.subtypes.end();
	    ++subtype) {
		log(debug,"Adds subtype");
		error = avahi_entry_group_add_service_subtype(d_group,
		                                              ToAvahiInterface(service.interface),
		                                              ToAvahiProtocol(service.protocol),
		                                              (AvahiPublishFlags)0,
		                                              d_name.get(),
		                                              type,
		                                              domain,
		                                              subtype->c_str());
		if(error) {
			std::ostringstream os;
			os << "avahi_entry_group_add_service_subtype() returned with error ( " << error
			   << ": " << avahi_strerror(error) << " )";
			avahi_simple_poll_quit(d_poll.get());
			throw std::runtime_error(os.str());
		}
	}
}

void AvahiPublisher::EntryGroupCallback(AvahiEntryGroup * g,
                                        AvahiEntryGroupState state,
                                        void * userdata) {
	AvahiPublisher * itself = reinterpret_cast<AvahiPublisher*>(userdata);
	assert(g == itself->d_group|| itself->d_group == NULL);
	if(itself->d_group == NULL) {
		itself->d_group = g;
	}
	/* Called whenever the entry group state changes */

	switch (state) {
	case AVAHI_ENTRY_GROUP_ESTABLISHED :
		/* The entry group has been established successfully */
		log(debug,"Service successfully established.");
		break;

	case AVAHI_ENTRY_GROUP_COLLISION : {
		itself->d_name = StrPtr(avahi_alternative_service_name(itself->d_name.get()),
		                        &avahi_free);
		log(debug,"Renaming service (int entry callback) to ",itself->d_name.get());
		itself->CreateServices(avahi_entry_group_get_client(g));
		break;
	}

	case AVAHI_ENTRY_GROUP_FAILURE : {
		int error = avahi_client_errno(avahi_entry_group_get_client(g));
		std::ostringstream os;
		os << "Entry group failure ( " << error << ": " << avahi_strerror(error) << " )";
		avahi_simple_poll_quit(itself->d_poll.get());
		throw std::runtime_error(os.str());
	}
	case AVAHI_ENTRY_GROUP_UNCOMMITED:
	case AVAHI_ENTRY_GROUP_REGISTERING:
		break;
	default :
		throw std::logic_error("Unexpected AvahiEntryGroupState value");
	}
}

AvahiStringList * AvahiPublisher::ToAvahiStrList(const Publisher::ListOfTxtDataFields & txtData) {
	AvahiStringList * list = NULL;

	for(Publisher::ListOfTxtDataFields::const_iterator td = txtData.begin();
	    td != txtData.end();
	    ++td) {
		list = avahi_string_list_add(list,td->c_str());
	}

	return list;
}

AvahiIfIndex AvahiPublisher::ToAvahiInterface(Interface i) {
	if(i == net::dnssd::IF_UNSPEC) {
		return AVAHI_IF_UNSPEC;
	}
	return i;
}


AvahiProtocol AvahiPublisher::ToAvahiProtocol(Protocol p) {
	switch (p){
	case PROTO_UNSPEC :
		return AVAHI_PROTO_UNSPEC;
	case PROTO_INET :
		return AVAHI_PROTO_INET;
	case PROTO_INET6 :
		return AVAHI_PROTO_INET6;
	default :
		throw std::logic_error("Unexpected net::dnssd::Protocol value");
	}
}


} // namespace dnssd
} // namespace net
