/**
 * \file DnsSdPublisher.cpp
 *
 * \date Nov 28, 2013
 * \author Alexandre Tuleu
 */

#include <biorob-cpp/log/Logger.h>

#include "Publisher.h"

#include <string.h>

#include <dns_sd.h>

 #include <sys/select.h>
#include <unistd.h>

#include <map>
#include <sstream>
#include <biorob-cpp/os/SystemError.h>

namespace net {

namespace dnssd {

class DnsSdPublisher :  public PublisherPrimpl {
public :
	DnsSdPublisher(Publisher & publisher);
	virtual ~DnsSdPublisher();

	void EnterMainLoop();
private :
	typedef std::tr1::shared_ptr<DNSServiceRef> ServiceRefPtr;
	typedef std::map<int,ServiceRefPtr> ServiceRefsByFD;

	static void DNSServiceRegisterReply(DNSServiceRef sdRef,
	                                    DNSServiceFlags flags,
	                                    DNSServiceErrorType errorCode,
	                                    const char *name,
	                                    const char *regtype,
	                                    const char *domain,
	                                    void *context );

	static std::pair<uint16_t,char*> EncodeTxtData(const Publisher::ListOfTxtDataFields & fields);
	static const std::string & NamedDNSErrorType(int error);
	ServiceRefsByFD d_servicesRefs;
};


PublisherPrimpl::Ptr PublisherPrimpl::Create(Publisher & publisher) {
	return PublisherPrimpl::Ptr(new DnsSdPublisher(publisher));
}


DnsSdPublisher::DnsSdPublisher(Publisher & publisher)
	: PublisherPrimpl(publisher){
}

DnsSdPublisher::~DnsSdPublisher() {
}

void DeallocateRef(DNSServiceRef * ref);

void DeallocateRef(DNSServiceRef * ref) {
	DNSServiceRefDeallocate(*ref);
}

#define wrap_str(str,name) \
	const char * name  = str.empty() ? NULL : str.c_str()

void DnsSdPublisher::EnterMainLoop() {
	for (ListOfServices::const_iterator service = Services().begin();
	     service != Services().end();
	     ++service) {
		DNSServiceRef * ref = new DNSServiceRef;

		wrap_str(service->type, type);
		wrap_str(service->domain, domain);
		wrap_str(service->host, host);
		std::pair<uint16_t,char*> encodedTxtData = EncodeTxtData(service->txtData);

		DNSServiceErrorType err = DNSServiceRegister(ref,
		                                             (DNSServiceFlags)0,
		                                             0,
		                                             GetPublisher().Name().c_str(),
		                                             type,
		                                             domain,
		                                             host,
		                                             service->port,
		                                             encodedTxtData.first,
		                                             encodedTxtData.second,
		                                             &DnsSdPublisher::DNSServiceRegisterReply,
		                                             reinterpret_cast<void*>(this));


		if(err != kDNSServiceErr_NoError ) {
			std::ostringstream os;
			os << "DNSServiceRegister(ref,0,0," << GetPublisher().Name()
			   << "," << (type == NULL ? "NULL" : type)
			   << "," << (domain == NULL ? "NULL" : domain)
			   << "," << (host == NULL ? "NULL" : host)
			   << "," << service->port
			   << "," << (int)encodedTxtData.first
			   << "," << encodedTxtData.second
			   << "," << "&DnsSdPublicher::DNSServiceRegisterReply"
			   << "," << this
			   <<") returned error:" << NamedDNSErrorType(err);

			delete []encodedTxtData.second;
			throw std::runtime_error(os.str());
		}
		//if succeed we defer to a smart pointer the deallocation of the service
		delete []encodedTxtData.second;
		ServiceRefPtr refPtr(ref, &DeallocateRef);

		//get the socket descriptor
		int fd = DNSServiceRefSockFD(*refPtr.get());
		if(fd == -1 ) {

			throw std::runtime_error("DNServiceRefSockFD() returned invalid socket.");
		}
		//everything is green, we just save the ref.
		d_servicesRefs[fd] = refPtr;
	}


	//enter the actual inifinite loop
	while(true) {
		//puts all the fd in the read set
		fd_set readSet,writeSet,exceptSet;
		FD_ZERO(&readSet);
		FD_ZERO(&writeSet);
		FD_ZERO(&exceptSet);
		int nfds = -1;
		for(ServiceRefsByFD::const_iterator s = d_servicesRefs.begin();
		    s != d_servicesRefs.end();
		    ++s) {
			nfds = std::max(nfds,s->first);
			FD_SET(s->first,&readSet);
		}

		//select all the FD that are ready to read.
		int selectRes = select(nfds+1,&readSet,&writeSet,&exceptSet,NULL);

		if(selectRes < 0) {
			//error has occured !!
			if( errno == EINTR ) {
				//simply a signal was caught : we exit garcefully.
				break;
			}
			//otherwise we notify the error through os::SystemError
			throw os::SystemError("select() returned with error",errno);
		}

		//process all the Service needed
		for(ServiceRefsByFD::const_iterator s = d_servicesRefs.begin();
		    s != d_servicesRefs.end();
		    ++s) {
			if (FD_ISSET(s->first,&readSet)) {
				DNSServiceProcessResult(*s->second);
			}
		}
	}
}



const std::string & DnsSdPublisher::NamedDNSErrorType(int error) {
	typedef std::map<int,std::string> ErrorNames;
	static ErrorNames errors;
	if(errors.empty()) {
#define add_error(errName) do{ \
	errors[errName] = #errName; \
}while(0)
		add_error(kDNSServiceErr_NoError);
		add_error(kDNSServiceErr_Unknown);
		add_error(kDNSServiceErr_NoSuchName);
		add_error(kDNSServiceErr_NoMemory);
		add_error(kDNSServiceErr_BadParam);
		add_error(kDNSServiceErr_BadReference);
		add_error(kDNSServiceErr_BadState);
		add_error(kDNSServiceErr_BadFlags);
		add_error(kDNSServiceErr_Unsupported);
		add_error(kDNSServiceErr_NotInitialized);
		add_error(kDNSServiceErr_AlreadyRegistered);
		add_error(kDNSServiceErr_NameConflict);
		add_error(kDNSServiceErr_Invalid);
		add_error(kDNSServiceErr_Firewall);
		add_error(kDNSServiceErr_Incompatible);
		add_error(kDNSServiceErr_BadInterfaceIndex);
		add_error(kDNSServiceErr_Refused);
		add_error(kDNSServiceErr_NoSuchRecord);
		add_error(kDNSServiceErr_NoAuth);
		add_error(kDNSServiceErr_NoSuchKey);
		add_error(kDNSServiceErr_NATTraversal);
		add_error(kDNSServiceErr_DoubleNAT);
		add_error(kDNSServiceErr_BadTime);
		/*add_error(kDNSServiceErr_BadSig);
		add_error(kDNSServiceErr_BadKey);
		add_error(kDNSServiceErr_Transient);
		add_error(kDNSServiceErr_ServiceNotRunning);
		add_error(kDNSServiceErr_NATPortMappingUnsupported);
		add_error(kDNSServiceErr_NATPortMappingDisabled);
		add_error(kDNSServiceErr_NoRouter);
		add_error(kDNSServiceErr_PollingMode);
		add_error(kDNSServiceErr_Timeout);*/
#undef add_error
	}
	ErrorNames::const_iterator fi = errors.find(error);
	if(fi == errors.end() ){
		return errors[kDNSServiceErr_Unknown];
	}

	return fi->second;
}

std::pair<uint16_t,char*> DnsSdPublisher::EncodeTxtData(const Publisher::ListOfTxtDataFields & fields) {
	std::pair<uint16_t,char*> res;

	std::vector<char> encodedString;


	for(Publisher::ListOfTxtDataFields::const_iterator f = fields.begin();
	    f != fields.end();
	    ++f) {
		if( f->size() > 255) {
			std::ostringstream os;
			os << "DNS TXT field '" << *f << "' has a unallowed size of " << f->size() << " 255 is the maximum";
			throw std::runtime_error(os.str());
		}
		encodedString.push_back((char)f->size());
		// Do not append the last NULL byte !
		for(std::string::size_type i = 0; i < f->size(); ++i) {
			encodedString.insert(encodedString.end(),f->at(i));
		}

	}

	res.first = encodedString.size();
	if(res.first == 0) {
		res.second = NULL;
	} else {
		char * buffer = new char[encodedString.size()];
		memcpy(buffer,&(encodedString[0]),encodedString.size());
	}


	return res;
}

void DnsSdPublisher::DNSServiceRegisterReply(DNSServiceRef /* sdRef */,
                                             DNSServiceFlags flags,
                                             DNSServiceErrorType errorCode,
                                             const char *name,
                                             const char *regtype,
                                             const char *domain,
                                             void * /* context */ ) {
	if(errorCode != kDNSServiceErr_NoError) {
		std::ostringstream os;
		os << "Registration of Service {name:" << name
		   << " type:" << regtype
		   << " domain:" << domain
		   << "} returned error(" << errorCode
		   << ": " << NamedDNSErrorType(errorCode);
		throw std::runtime_error(os.str());
	}

	if(! (flags & kDNSServiceFlagsAdd) ) {
		std::ostringstream os;
		os << "Got name conflict resolution for Service {name:" << name
		   << " type:" << regtype
		   << " domain:" << domain
		   << "}";
		throw std::runtime_error(os.str());
	}

	log(debug, "Successfuly registered Service {name:",name," type:",regtype," domain:",domain,"}.");
}

} /* namespace dnssd */

} /* namespace net */
