/*
 * \file CliArgParser.h
 *
 *  \date Oct 10, 2013
 *  \author Alexandre Tuleu
 */

#ifndef BIOROB_CPP_CLIARGPARSER_H_
#define BIOROB_CPP_CLIARGPARSER_H_


#include <string>
#include <vector>
#include <map>
#include <tr1/memory>
#include <set>

namespace options {

class CliArgParser {
public:
	CliArgParser(const std::string & description);

	void AppendUsage(const std::string & usage);

	void EnableHelp(bool on=true);
	bool HelpEnabled() const;

	void IgnoreUnusedArguments(bool on=true);
	bool UnusedArgumentsIgnored() const;

	void AddOption(const std::string & name,
	               bool & value,
	               const std::string & description,
	               char shortName = 0);

	template<typename T> void AddParameter(const std::string & name,
	                                       T & value,
	                                       const std::string & description,
	                                       bool mandatory = false,
	                                       char shortName = 0);

	template<typename T> void AddListParameter(const std::string & name,
	                                           typename std::vector<T> & value,
	                                           const std::string & decription,
	                                           char shortName = 0);

	enum ActionOnError {
		EXIT_ON_ERROR,
		THROW_ON_ERROR
	};


	std::vector<std::string> Parse(int argc,
	                               char ** argv,
	                               ActionOnError = EXIT_ON_ERROR);

	void PrintHelp(std::ostream & out) const;

private :
	class ValueHolderBase {
	public :
		ValueHolderBase(const std::string & name,
		                const std::string & description,
		                char shortName,
		                bool mandatory,
		                int maxExpectedArgs);
		virtual ~ValueHolderBase();
		typedef std::tr1::shared_ptr<ValueHolderBase> Ptr;
		virtual void ConsumeArg(const char * arg) = 0;
		virtual void PrintDescription(std::ostream & out) const = 0;

		const std::string & Name() const;
		const std::string & Description() const;
		std::string FullName() const;

		bool CheckMandatory() const;
		

		int MaxExpectedArgs() const;

	protected :
		void MarkModified();
		bool Modified() const;
	private :
		std::string d_name;
		std::string d_description;
		char d_shortName;
		bool d_mandatory;
		bool d_modified;
		int d_maxExpectedArgs;

	};



	template <typename T>
	class ValueHolder : public ValueHolderBase {
	public :
		ValueHolder(const std::string & name,
                    const std::string & description,
                    char shortName,
                    bool mandatory,
                    T & value);
		virtual ~ValueHolder();
		void ConsumeArg(const char * arg);
		void PrintDescription(std::ostream & out) const;
	private :
		T & d_value;
	};

	template <typename T>
	class ListValueHolder : public ValueHolderBase {
	public :
		ListValueHolder(const std::string & name,
                        const std::string & description,
                        char shortName,
                        std::vector<T> & values);
		virtual ~ListValueHolder();
		void ConsumeArg(const char * arg);
		void PrintDescription(std::ostream & out) const;
	private :
		std::vector<T> & d_values;

	};

	typedef std::map<char,ValueHolderBase::Ptr> ValuesByShort;
	typedef std::map<std::string,ValueHolderBase::Ptr> ValuesByLong;


	void ParseOrExit(int argc, char ** argv, std::vector<std::string> & values);
	void ParseOrThrow(int argc, char ** argv, std::vector<std::string> & values);
	
	void ParseShortOption(const std::string & name, 
	                      const std::string & inlineValue);

	void CheckOptionName(const std::string & name,char shortName) const;
	static void CheckOptionNameCorrectness(const std::string & name,
	                                       char shortName);
	void CheckOptionNameAvailability(const std::string & name,
	                                 char shortName) const;

	void ExtractOptionName(const char * argv, 
	                       std::vector<std::string> & optionName,
	                       std::vector<std::string> & inlinedValue);

	void PrintHelpAndExitIfEnabled() const;

	bool d_helpEnabled;
	bool d_ignoreUnusedArgs;

	std::string d_description;
	std::vector<std::string> d_usages;

	ValuesByLong d_longValues;
	ValuesByShort d_shortValues;

};

} /* namespace options */


#endif /* BIOROB_CPP_CLIARGPARSER_H_ */

#include "CliArgParser.imp.h"
