/*
 * \file CliArgumentParserUTest.cpp
 *
 *  \date Oct 10, 2013
 *  \author Alexandre Tuleu
 */

#include "CliArgParserUTest.h"
#include <stdexcept>
#include <cctype>
#include <algorithm>

#define EXPECT_NO_STDEXCEPT(statement) do { \
	try { \
	statement; \
	} catch (const std::exception & e) { \
		FAIL() << "Expected: " << #statement << " not to throw exception." << std::endl \
		       << "  Actually: it throws with message '" << e.what() << "'"; \
	} catch (...) { \
		FAIL() << "Expected: " << #statement << " not to throw exception." << std::endl \
		       << "  Actually: it does."; \
	} \
}while(0)


CLIArgsSplitter::CLIArgsSplitter(const std::string & cli)
	: d_argc(0)
	, d_argv(NULL){
	Init(cli);
}

CLIArgsSplitter::~CLIArgsSplitter(){
	if(d_argv == NULL ) {
		return;
	}

	for(int i = 0;
	    i < d_argc;
	    ++i) {
		if(d_argv[i] != NULL) {
			delete[] d_argv[i];
		}
	}

	delete[] d_argv;
}

int CLIArgsSplitter::Argc(){
	return d_argc;
}

char ** CLIArgsSplitter::Argv(){
	return d_argv;
}

void CLIArgsSplitter::Init(const std::string & cli) {
	/* split the string with space or "' */
	std::vector<std::string> token;
	std::string::size_type current = 0;


	while(current != std::string::npos) {
		if(cli[current] == '"' || cli[current] == '\'') {
			std::string::size_type next = cli.find_first_of(cli[current],
			                                                current + 1);
			if(next == std::string::npos || current + 1 == cli.size() ) {
				throw std::runtime_error("Unmatched \" or ' in '" + cli + "'");
			}
			token.push_back(cli.substr(current + 1, next - current - 1));
			if (next + 1 == cli.size()) {
				break;
			}
			current = cli.find_first_not_of(" ",next + 1);
			continue;
		}

		std::string::size_type next = cli.find_first_of(" ",current);
		if(next == std::string::npos) {
			token.push_back(cli.substr(current));
			break;
		}
		token.push_back(cli.substr(current,next - current));
		current = cli.find_first_not_of(" ",next);
	}



	//convert to C like
	d_argc = token.size();
	d_argv = new char*[d_argc];
	for(unsigned int i = 0; i < token.size(); ++i) {
		d_argv[i] = new char[token[i].size() + 1];
		memcpy(d_argv[i],&(token[i][0]),token[i].size());
		d_argv[i][token[i].size()] = 0;
	}
}


std::string CliArgParserUTest::s_fooCommandDescription("foo-command is a command that does blah things.");

void CliArgParserUTest::SetUp(){
}

void CliArgParserUTest::TearDown() {
}


TEST(CliArgSplitter,WorksAsBash) {
	CLIArgsSplitter args("this is a \"really complex\" 'test case'");
	ASSERT_EQ(5,args.Argc());
	EXPECT_STREQ("this",args.Argv()[0]);
	EXPECT_STREQ("is",args.Argv()[1]);
	EXPECT_STREQ("a",args.Argv()[2]);
	EXPECT_STREQ("really complex",args.Argv()[3]);
	EXPECT_STREQ("test case",args.Argv()[4]);

	EXPECT_THROW(CLIArgsSplitter bad("this is not \"parsable "),std::runtime_error);
	EXPECT_THROW(CLIArgsSplitter bad("this is not 'parsable "),std::runtime_error);
}

TEST_F(CliArgParserUTest,ParseInteger) {
	ParseOneParameter<int>(0,"-42");
	ParseOneParameter<unsigned int>(0,"42");

	ParseOneParameter<long>(0,"-6942");
	ParseOneParameter<unsigned long>(0,"6942");
}

TEST_F(CliArgParserUTest,ParseSimpleString) {
	ParseOneParameter<std::string>("foo","bar");
}

TEST_F(CliArgParserUTest,ParsesFloat) {
	double value(12345.6789);
	double parsedValue(0.0);
	std::ostringstream os;
	os <<"foo-command --foo-option "
	   << std::scientific << std::setprecision(10) << value;
	CLIArgsSplitter args(os.str());
	options::CliArgParser p(s_fooCommandDescription);

	p.AddParameter("foo-option",parsedValue,"any description");

	EXPECT_NO_THROW(p.Parse(args.Argc(),
	                        args.Argv(),
	                        options::CliArgParser::THROW_ON_ERROR));

	EXPECT_FLOAT_EQ(value,parsedValue);

}

TEST_F(CliArgParserUTest,ParseComplexString) {
	std::string longValue("this is a long value");
	CLIArgsSplitter args("foo-command --option \"" + longValue + "\"");

	options::CliArgParser p(s_fooCommandDescription);
	std::string value;
	p.AddParameter<std::string>("option",value,"blah option");

	EXPECT_TRUE(value.empty());

	EXPECT_NO_THROW(p.Parse(args.Argc(),
	                        args.Argv(),
	                        options::CliArgParser::THROW_ON_ERROR));

	EXPECT_EQ(longValue,value);

}


TEST_F(CliArgParserUTest,ParsesWithMultipleFormat) {
	CLIArgsSplitter args("foo-command --aoption foo -b bar --coption=baz -d=foobar");
	options::CliArgParser p(s_fooCommandDescription);
	std::string param[4];
	p.AddParameter("aoption",param[0],"any description",false,'a');
	p.AddParameter("boption",param[1],"any description",false,'b');
	p.AddParameter("coption",param[2],"any description",false,'c');
	p.AddParameter("doption",param[3],"any description",false,'d');

	EXPECT_NO_STDEXCEPT(p.Parse(args.Argc(),
	                            args.Argv(),
	                            options::CliArgParser::THROW_ON_ERROR));
	EXPECT_EQ("foo",param[0]);
	EXPECT_EQ("bar",param[1]);
	EXPECT_EQ("baz",param[2]);
	EXPECT_EQ("foobar",param[3]);
}


TEST_F(CliArgParserUTest,AcceptsOnlyUniqueName) {
	options::CliArgParser p(s_fooCommandDescription);
	bool value;
	EXPECT_NO_THROW(p.AddOption("option",value,"any description",'o'));
	EXPECT_THROW(p.AddOption("foo-option",
	                         value,
	                         "any description",'o'),
	             std::logic_error);
	EXPECT_THROW(p.AddOption("option",
	                         value,
	                         "any description"),
	             std::logic_error);

	unsigned int param;
	EXPECT_NO_THROW(p.AddParameter<unsigned int>("param",
	                                             param,
	                                             "any description",
	                                             false,
	                                             'p'));
	EXPECT_THROW(p.AddParameter<unsigned int>("param",
	                                          param,
	                                          "any descriptiion",
	                                          false),
	             std::logic_error);
	EXPECT_THROW(p.AddParameter<unsigned int>("foo-param",
	                                          param,
	                                          "any descriptiion",
	                                          false,
	                                          'p'),
	             std::logic_error);

}


TEST_F(CliArgParserUTest,RefusesEmptyDescription) {
	EXPECT_THROW(options::CliArgParser bad(""),std::logic_error);
	options::CliArgParser *p;
	EXPECT_NO_THROW(p = new options::CliArgParser(s_fooCommandDescription));
	bool opt;
	int param;
	std::vector<int> params;
	EXPECT_THROW(p->AddOption("option",opt,""),std::logic_error);
	EXPECT_THROW(p->AddParameter<int>("param",param,""),std::logic_error);
	EXPECT_THROW(p->AddListParameter<int>("params",params,""),std::logic_error);
}

TEST_F(CliArgParserUTest,ChecksProgamNamePresence) {
	options::CliArgParser p(s_fooCommandDescription);
	CLIArgsSplitter goodArgs("foo-command");
	EXPECT_NO_THROW(p.Parse(goodArgs.Argc(),
	                        goodArgs.Argv(),
	                        options::CliArgParser::THROW_ON_ERROR));
	CLIArgsSplitter badArgs("");
	EXPECT_THROW(p.Parse(badArgs.Argc(),
	                     badArgs.Argv(),
	                     options::CliArgParser::THROW_ON_ERROR),
	             std::logic_error);

}

TEST_F(CliArgParserUTest, CatchesUnparsedValues) {
	CLIArgsSplitter args("foo-command --option foo filename1 filename2");
	options::CliArgParser p(s_fooCommandDescription);
	std::string param;
	p.AddParameter("option",param,"any description");
	std::vector<std::string> remainingArgs;
	EXPECT_NO_THROW(remainingArgs = p.Parse(args.Argc(),
	                                        args.Argv(),
	                                        options::CliArgParser::THROW_ON_ERROR));
	ASSERT_EQ(2,remainingArgs.size());
	EXPECT_EQ("filename1",remainingArgs[0]);
	EXPECT_EQ("filename2",remainingArgs[1]);

}

TEST_F(CliArgParserUTest, SetsOptionsByDefaultToFalse) {
	CLIArgsSplitter args("foo-command -a");

	bool opt = true;

	options::CliArgParser p(s_fooCommandDescription);
	EXPECT_TRUE(opt);
	p.AddOption("optiona",opt,"any description",'a');
	EXPECT_FALSE(opt) << "Option holder does not set it to false";
	EXPECT_NO_THROW(p.Parse(args.Argc(),
	                        args.Argv(),
	                        options::CliArgParser::THROW_ON_ERROR));
	EXPECT_TRUE(opt);
}

TEST_F(CliArgParserUTest, CatchesMultipleOptionFormat) {
	CLIArgsSplitter args("foo-command --optione -DBc -a");

	bool opts[5];

	options::CliArgParser p(s_fooCommandDescription);
	p.AddOption("optiona",opts[0],"any description",'a');
	p.AddOption("optionb",opts[1],"any description",'B');
	p.AddOption("optionc",opts[2],"any description",'c');
	p.AddOption("optiond",opts[3],"any description",'D');
	p.AddOption("optione",opts[4],"any description",'e');

	EXPECT_NO_THROW(p.Parse(args.Argc(),
	                        args.Argv(),
	                        options::CliArgParser::THROW_ON_ERROR));

	for(unsigned int i = 0 ; i < 5 ; ++i) {
		EXPECT_TRUE(opts[i])  << "for index " << i;
	}

}


TEST_F(CliArgParserUTest, ThrowsOnbadOptionParse) {
	CLIArgsSplitter args("foo-command --a=trucQuiExistePas");

	bool opt;

	options::CliArgParser p(s_fooCommandDescription);
	p.AddOption("optiona",opt,"any description",'a');
	EXPECT_THROW(p.Parse(args.Argc(),
	                     args.Argv(),
	                     options::CliArgParser::THROW_ON_ERROR),
	             std::runtime_error);

}


TEST_F(CliArgParserUTest, ForbidsBadOptionNames) {
	options::CliArgParser p(s_fooCommandDescription);

	std::vector<std::string> badOptionName;
	std::string goodOptionName("foo-option");
	std::string badShortName = "0123456789!&%*-+";

	badOptionName.push_back("1foo");
	badOptionName.push_back("foo bar");
	badOptionName.push_back("!foo");
	badOptionName.push_back("foo+bar");

	for(std::vector<std::string>::const_iterator bon = badOptionName.begin();
	    bon != badOptionName.end();
	    ++bon) {
		bool opt;
		std::string param;
		std::vector<std::string> params;
		EXPECT_THROW(p.AddOption(*bon,opt,"any description"),std::logic_error) << "with option name '" << *bon << "'";
		EXPECT_THROW(p.AddParameter(*bon,param,"any description"),std::logic_error) << "with option name '" << *bon << "'";
		EXPECT_THROW(p.AddListParameter(*bon,params,"any description"),std::logic_error) << "with option name '" << *bon << "'";
	}


	for(std::string::iterator c = badShortName.begin();
	    c != badShortName.end();
	    ++c) {
				bool opt;
		std::string param;
		std::vector<std::string> params;
		EXPECT_THROW(p.AddOption(goodOptionName,opt,"any description",*c),std::logic_error) << "with option short name '" << *c << "'";
		EXPECT_THROW(p.AddParameter(goodOptionName,param,"any description",false,*c),std::logic_error) << "with option short name '" << *c << "'";
		EXPECT_THROW(p.AddListParameter(goodOptionName,params,"any description",*c),std::logic_error) << "with option short name '" << *c << "'";
	}
	std::string param;
	try {
		p.AddParameter("foo-option",param,"any description");
		p.AddParameter("foo_option",param,"any description");
	} catch ( const std::exception & e) {
		FAIL() << "Got exception : " << e.what();
	}

}



TEST_F(CliArgParserUTest, ParsesListWithMultipleFormat) {
	CLIArgsSplitter args("foo-progam --optiona foo --optiona bar --optiona baz --optionb foobar foobaz foofoo -C=barfoo -C=bazfoo -C=barbar");
	options::CliArgParser p(s_fooCommandDescription);

	std::vector<std::string> params1,params2,params3;

	p.AddListParameter("optiona",params1,"any description",'A');
	p.AddListParameter("optionb",params2,"any description",'B');
	p.AddListParameter("optionc",params3,"any description",'C');

	EXPECT_NO_STDEXCEPT(p.Parse(args.Argc(),
	                            args.Argv(),
	                            options::CliArgParser::THROW_ON_ERROR));


	/* It also test if order is preserved ! */
	ASSERT_EQ(3,params1.size());
	ASSERT_EQ(3,params2.size());
	ASSERT_EQ(3,params3.size());

	EXPECT_EQ("foo",params1[0]);
	EXPECT_EQ("bar",params1[1]);
	EXPECT_EQ("baz",params1[2]);

	EXPECT_EQ("foobar",params2[0]);
	EXPECT_EQ("foobaz",params2[1]);
	EXPECT_EQ("foofoo",params2[2]);

	EXPECT_EQ("barfoo",params3[0]);
	EXPECT_EQ("bazfoo",params3[1]);
	EXPECT_EQ("barbar",params3[2]);
}


TEST_F(CliArgParserUTest, ThrowsOnRepetetiveParameter) {
	CLIArgsSplitter args("foo-command --foo-param value1 --foo-param value2");

	std::string param;

	options::CliArgParser p(s_fooCommandDescription);
	p.AddParameter("foo-option",param,"any description");

	EXPECT_THROW(p.Parse(args.Argc(),
	                     args.Argv(),
	                     options::CliArgParser::THROW_ON_ERROR),
	             std::runtime_error);

}


TEST_F(CliArgParserUTest, PrintsHelfpfulMessage) {
	std::string description("foo-command : a command that does blah thing");
	std::string blahOptionDesc("blah thing to enable");
	std::string chombierParamDesc("chombier thing to tune");
	std::string awesoneParamsList("awesome parameter to add");
	std::string firstUsage ("foo-command [-c/--chombier value] file");
	std::string secondUsage ("foo-command [-a file1 ... -a file2]");

	bool blahEnabled;
	std::string chombier("CHOMBIER!!");
	std::vector<std::string> awesome;

	std::ostringstream expected;
	expected << description << std::endl
	         << "usage:" << std::endl
	         << "  " << firstUsage << std::endl
	         << "  " << secondUsage << std::endl
	         << "arguments:" << std::endl
	         << "  --help/-h : Prints this help message and exits."
	         << std::endl
	         << "  --awesome/-a : " << awesoneParamsList
	         << " (list of values)" << std::endl
	         << "  --blah/-b : " << blahOptionDesc
	         << " (default: disabled)" << std::endl
	         << "  --chombier/-c : " << chombierParamDesc
	         << " (default: " << chombier << ")" << std::endl;

	options::CliArgParser p(description);

	p.AppendUsage(firstUsage);
	p.AppendUsage(secondUsage);
	p.EnableHelp(true);
	p.AddOption("blah",blahEnabled,blahOptionDesc,'b');
	p.AddParameter("chombier",chombier,chombierParamDesc,false,'c');
	p.AddListParameter("awesome",awesome,awesoneParamsList,'a');
	std::ostringstream printed;
	p.PrintHelp(printed);

	EXPECT_EQ(expected.str(),printed.str());
}


TEST_F(CliArgParserUTest,ReservesHelpArgument) {
	options::CliArgParser p(s_fooCommandDescription);
	//even if help is disabled...
	p.EnableHelp(false);
	bool foo;
	EXPECT_THROW(p.AddOption("help",foo,"any description"),std::logic_error);
	EXPECT_THROW(p.AddOption("foo-option",foo,"any description",'h'),
	             std::logic_error);

}


TEST_F(CliArgParserUTest,PrintsHelpAndExists) {
	CLIArgsSplitter args("foo-command --help");
	options::CliArgParser p(s_fooCommandDescription);

	p.EnableHelp(false);

	//nothings happens
	EXPECT_NO_THROW(p.Parse(args.Argc(),
	                        args.Argv(),
	                        options::CliArgParser::THROW_ON_ERROR));
	p.EnableHelp(true);
	EXPECT_EXIT(p.Parse(args.Argc(),
	                        args.Argv(),
	                        options::CliArgParser::THROW_ON_ERROR),
	            ::testing::ExitedWithCode(0),
	            "");
}

