/*
 * Copyright (c) 2012 - 2013 Alexandre Tuleu
 *
 * This file is part of biorob-cpp.
 *
 *   biorob-cpp is free software: you can redistribute it and/or modify
 *   it under the terms of the GNU Lesser General Public License as
 *   published by the Free Software Foundation, either version 3 of the
 *   License, or (at your option) any later version.
 *
 *   biorob-cpp is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of both the GNU General Public
 *   License AND GNU Lesser General Public License along with
 *   biorob-cpp.  If not, see <http://www.gnu.org/licenses/>.
 */


/**
 * Pathname.cpp
 *
 *  Created on: 7 nov. 2012
 *      Author: akay
 */

#include "Pathname.h"

#include <algorithm>

#include <iostream>

using namespace os;

Pathname::Pathname() : d_computed(NONE), d_pathname(),d_path(),d_filename(),d_basename(), d_suffix() {}

Pathname::Pathname(const std::string & name) : d_computed(NONE), d_pathname(name),d_path(),d_filename(),d_basename(), d_suffix(){}

Pathname::~Pathname(){}

void Pathname::SetPathname(const std::string & name){
	d_pathname = name;
	d_computed = NONE;
}

const std::string & Pathname::CompletePath() const {
	return d_pathname;
}

bool Pathname::IsAbsolute() const {
	if(d_pathname.empty()){
		return false;
	}
	return d_pathname.at(0) == '/';
}

bool Pathname::IsRelative() const {
	if(d_pathname.empty()){
		return false;
	}

	return d_pathname.at(0) != '/';
}


const std::string & Pathname::Filename(){
	if(d_computed & FILENAME_AND_PATH){
		return d_filename;
	}
	ComputeFilenameAndPath();
	return d_filename;
}

const std::string & Pathname::Path(){
	if(d_computed & FILENAME_AND_PATH){
		return d_path;
	}
	ComputeFilenameAndPath();
	return d_path;
}

void Pathname::ComputeFilenameAndPath(){
	d_computed |= FILENAME_AND_PATH;

	if(d_pathname.empty()){
		d_filename.clear();
		d_path.clear();
		return;
	}

	std::string::size_type fend = d_pathname.size();

	if(d_pathname.at(fend - 1) == '/'){
		if(d_pathname.size() == 1){
			d_filename.clear();
			d_path = d_pathname;
			return;
		}
		--fend;
	}

	std::string::size_type fbegin = fend - 1;

	while(d_pathname.at(fbegin) != '/' && fbegin != 0){
		--fbegin;
	}
	if(d_pathname.at(fbegin) == '/'){
		++fbegin;
	}


	d_filename = d_pathname.substr(fbegin,fend-fbegin);
	d_path = d_pathname.substr(0,fbegin);

}


const std::string & Pathname::Basename(){
	if(d_computed & BASENAME_AND_SUFFIX){
		return d_basename;
	}
	const_cast<Pathname*>(this)->ComputeBaseAndSuffix();
	return d_basename;
}

const std::string & Pathname::Suffix(){
	if(d_computed & BASENAME_AND_SUFFIX){
		return d_suffix;
	}
	const_cast<Pathname*>(this)->ComputeBaseAndSuffix();
	return d_suffix;
}

void Pathname::ComputeBaseAndSuffix(){
	if(!(d_computed & FILENAME_AND_PATH) ){
		ComputeFilenameAndPath();
	}
	d_computed |= BASENAME_AND_SUFFIX;
	if(d_filename.empty()){
		d_basename.clear();
		d_suffix.clear();
		return;
	}

	std::string::size_type lastPoint = d_filename.size() - 1;

	while(d_filename.at(lastPoint) != '.' && lastPoint != 0){
		--lastPoint;
	}

	if(lastPoint <= 1){
		d_basename = d_filename;
		d_suffix.clear();
		return;
	}

	d_basename = d_filename.substr(0,lastPoint);
	d_suffix = d_filename.substr(lastPoint + 1);

}

bool Pathname::HasPrefix(const std::string & prefix){
	if(d_pathname.empty()){
		return false;
	}
	return Filename().substr(0,prefix.size()) == prefix;
}
